document.addEventListener("DOMContentLoaded", function() {
    const videoPlayer = videojs('video-player', {
        fluid: true,
        controls: true,
        autoplay: true,
        preload: 'auto'
    });

    // Ensure the player doesn't pause when toggling fullscreen
    videoPlayer.on('fullscreenchange', function() {
        if (videoPlayer.isFullscreen()) {
            // Ensure the player remains playing when fullscreen mode is toggled
            videoPlayer.play().catch(error => console.error('Error resuming playback:', error));
        }
    });

    videoPlayer.on('dblclick', function() {
        if (screen.orientation && screen.orientation.lock) {
            screen.orientation.lock('landscape').catch(function(error) {
                console.error('Screen orientation lock failed:', error);
            });
        } else if (screen.lockOrientation) {
            screen.lockOrientation('landscape');
        } else if (screen.mozLockOrientation) {
            screen.mozLockOrientation('landscape');
        } else if (screen.msLockOrientation) {
            screen.msLockOrientation('landscape');
        } else if (typeof window.cordova !== 'undefined' && window.cordova.plugins.screenorientation) {
            window.cordova.plugins.screenorientation.lock('landscape');
        }
    });

    videoPlayer.on('click', function() {
        toggleFullscreen();
    });

    function toggleFullscreen() {
        if (!document.fullscreenElement) {
            if (videoPlayer.el().requestFullscreen) {
                videoPlayer.el().requestFullscreen();
            } else if (videoPlayer.el().mozRequestFullScreen) { // Firefox
                videoPlayer.el().mozRequestFullScreen();
            } else if (videoPlayer.el().webkitRequestFullscreen) { // Chrome, Safari and Opera
                videoPlayer.el().webkitRequestFullscreen();
            } else if (videoPlayer.el().msRequestFullscreen) { // IE/Edge
                videoPlayer.el().msRequestFullscreen();
            }
        } else {
            if (document.exitFullscreen) {
                document.exitFullscreen();
            } else if (document.mozCancelFullScreen) { // Firefox
                document.mozCancelFullScreen();
            } else if (document.webkitExitFullscreen) { // Chrome, Safari and Opera
                document.webkitExitFullscreen();
            } else if (document.msExitFullscreen) { // IE/Edge
                document.msExitFullscreen();
            }
        }
    }

    const progressBar = document.getElementById('progress-bar');
    let progress = 0;
    let interval = setInterval(function() {
        progress += Math.random() * 5;
        progressBar.style.width = progress + '%';

        if (progress >= 100) {
            clearInterval(interval);
            progressBar.style.width = '100%';
            setTimeout(function() {
                document.getElementById('progress-container').style.display = 'none';
            }, 500);
        }
    }, 100);

    document.addEventListener("scroll", function() {
        const playlistContainer = document.querySelector('.playlist-container');
        const rect = playlistContainer.getBoundingClientRect();
        const isVisible = (rect.top >= 0 && rect.bottom <= window.innerHeight);
        if (isVisible) {
            document.body.classList.add('playlist-visible');
        } else {
            document.body.classList.remove('playlist-visible');
        }
    });

    const playlistUrl = 'https://m3u.ch/pl/9ac3db698c37446a1fe2811fe93c8533_89831d2052d99038edd0275cddf93560.m3u';
    
    // Show loading message while fetching channels
    showLoadingMessage();
    fetchAndParsePlaylist(playlistUrl);

    function fetchAndParsePlaylist(playlistUrl) {
        fetch(playlistUrl)
            .then(response => response.text())
            .then(data => {
                const lines = data.trim().split('\n');
                const channels = extractChannels(lines);
                populatePlaylistAndTabs(channels);
                removeLoadingMessage();
                if (channels.length === 0) {
                    showNoChannelsMessage();
                }
            })
            .catch(error => {
                console.error('Error fetching playlist data:', error);
                removeLoadingMessage();
                showNoChannelsMessage();
            });
    }

    function extractChannels(lines) {
        const channels = [];
        let currentChannel = null;
        lines.forEach(line => {
            if (line.startsWith('#EXTINF:')) {
                const [, tvgNameMatch] = line.match(/tvg-name="(.*?)"/) || [];
                const [, logoUrl] = line.match(/tvg-logo="(.*?)"/) || [];
                const [, groupName] = line.match(/group-title="(.*?)"/) || [];
                const [, channelName] = line.match(/,(.*)/) || [];
                const name = tvgNameMatch || channelName;
                currentChannel = { name, logoUrl, groupName };
            } else if (line.trim() && currentChannel) {
                channels.push({ ...currentChannel, url: line.trim() });
                currentChannel = null;
            }
        });
        return channels;
    }

    function populatePlaylistAndTabs(channels) {
        const playlistContainer = document.getElementById('playlist');
        const groupTabsContainer = document.getElementById('group-tabs');
        const groupedChannels = groupChannels(channels);

        for (const groupName in groupedChannels) {
            const groupTab = document.createElement('div');
            groupTab.classList.add('group-tab');
            groupTab.textContent = groupName;
            groupTab.addEventListener('click', function() {
                document.querySelectorAll('.group-tab').forEach(tab => {
                    tab.classList.remove('active');
                });
                this.classList.add('active');
                
                document.querySelectorAll('.group-content').forEach(content => {
                    content.style.display = 'none';
                });
                const groupContent = document.getElementById(groupName.toLowerCase().replace(/\s+/g, '-'));
                groupContent.style.display = 'block';
            });
            groupTabsContainer.appendChild(groupTab);

            const groupContent = document.createElement('div');
            groupContent.classList.add('group-content');
            groupContent.id = groupName.toLowerCase().replace(/\s+/g, '-');

            groupedChannels[groupName].forEach(channel => {
                const playlistItem = createPlaylistItem(channel);
                groupContent.appendChild(playlistItem);
            });
            playlistContainer.appendChild(groupContent);

            if (groupName === Object.keys(groupedChannels)[0]) {
                groupTab.click();
            }
        }
    }

    function groupChannels(channels) {
        const groupedChannels = {};
        channels.forEach(channel => {
            const groupName = channel.groupName || 'Uncategorized';
            if (!groupedChannels[groupName]) {
                groupedChannels[groupName] = [];
            }
            groupedChannels[groupName].push(channel);
        });
        return groupedChannels;
    }

    function createPlaylistItem(channel) {
        const playlistItem = document.createElement('div');
        playlistItem.classList.add('channel');
        playlistItem.textContent = channel.name;

        if (channel.logoUrl) {
            const logoImage = document.createElement('img');
            logoImage.src = channel.logoUrl;
            logoImage.alt = channel.name;
            logoImage.classList.add('logo');
            playlistItem.appendChild(logoImage);
        }

        playlistItem.addEventListener('click', function() {
            videoPlayer.src(channel.url);
            updateChannelName(channel.name);
            updateCategoryName(channel.groupName);
        });

        return playlistItem;
    }

    function updateChannelName(name) {
        const channelNameElement = document.getElementById('channel-name');
        channelNameElement.textContent = name;
    }
    
    function updateCategoryName(name) {
        const categoryNameElement = document.getElementById('category-name');
        categoryNameElement.textContent = name;
    }

    function showLoadingMessage() {
        const playlistContainer = document.getElementById('playlist');
        const loadingMessage = document.createElement('div');
        loadingMessage.classList.add('loading-message');
        loadingMessage.textContent = 'Loading channels...';
        playlistContainer.appendChild(loadingMessage);
    }

    function removeLoadingMessage() {
        const loadingMessage = document.querySelector('.loading-message');
        if (loadingMessage) {
            loadingMessage.remove();
        }
    }

    function showNoChannelsMessage() {
        const playlistContainer = document.getElementById('playlist');
        const noChannelsMessage = document.createElement('div');
        noChannelsMessage.classList.add('no-channels-message');
        noChannelsMessage.textContent = 'No channels available.';
        playlistContainer.appendChild(noChannelsMessage);
    }

    function removeNoChannelsMessage() {
        const noChannelsMessage = document.querySelector('.no-channels-message');
        if (noChannelsMessage) {
            noChannelsMessage.remove();
        }
    }

    const searchBar = document.getElementById('search-bar');
    searchBar.addEventListener('input', function() {
        const searchValue = searchBar.value.toLowerCase();
        const channels = document.querySelectorAll('.channel');
        channels.forEach(channel => {
            const channelName = channel.textContent.toLowerCase();
            if (channelName.includes(searchValue)) {
                channel.style.display = 'block';
            } else {
                channel.style.display = 'none';
            }
        });

        const visibleChannels = Array.from(channels).filter(channel => channel.style.display === 'block');
        if (visibleChannels.length === 0) {
            showNoChannelsMessage();
        } else {
            removeNoChannelsMessage();
        }
    });

    // Function to clear session data
    function clearSessionData() {
        localStorage.removeItem('sessionData');
        console.log('Session data cleared');
    }

    // Clear session data on page unload or app pause
    window.onbeforeunload = function() {
        clearSessionData();
    };

    document.addEventListener("pause", clearSessionData, false);
});
// Firebase configuration
const firebaseConfig = {
    apiKey: "YOUR_API_KEY",
    authDomain: "YOUR_PROJECT_ID.firebaseapp.com",
    projectId: "YOUR_PROJECT_ID",
    storageBucket: "YOUR_PROJECT_ID.appspot.com",
    messagingSenderId: "YOUR_MESSAGING_SENDER_ID",
    appId: "YOUR_APP_ID"
};

// Initialize Firebase
firebase.initializeApp(firebaseConfig);

// Initialize Remote Config
const remoteConfig = firebase.remoteConfig();
remoteConfig.settings = {
    minimumFetchIntervalMillis: 3600000 // Fetch every hour
};

// Fetch and activate Remote Config values
remoteConfig.fetchAndActivate()
    .then(() => {
        const newVersionAvailable = remoteConfig.getString('new_version_available');
        if (newVersionAvailable === 'true') {
            addNotification('A new version of the app is available. Please update from the Play Store.');
        }
    })
    .catch(error => {
        console.error('Error fetching remote config:', error);
    });

// Initialize Firebase Cloud Messaging
const messaging = firebase.messaging();
messaging.onMessage(payload => {
    console.log('Message received. ', payload);
    addNotification(payload.notification.body);
});

function addNotification(message) {
    const notificationPanel = document.getElementById('notifications');
    const notification = document.createElement('div');
    notification.classList.add('notification');
    notification.textContent = message;
    notificationPanel.appendChild(notification);
}

function toggleNotifications() {
    const panel = document.getElementById('notification-panel');
    panel.style.display = (panel.style.display === 'none' || panel.style.display === '') ? 'block' : 'none';
}

// Request notification permission and get token
messaging.requestPermission()
    .then(() => {
        console.log('Notification permission granted.');
        return messaging.getToken();
    })
    .then(token => {
        console.log('FCM Token:', token);
        // Save token to your server to send notifications later
    })
    .catch(error => {
        console.error('Error getting notification permission:', error);
    });
